import pytest
from wandb.summary import FileSummary
from click.testing import CliRunner
from wandb import Histogram
import numpy as np
import os
import json
import torch
import tensorflow as tf
import pandas as pd


@pytest.fixture
def summary():
    with CliRunner().isolated_filesystem():
        s = FileSummary()
        s.update({"foo": "init"})
        yield s


def test_set_attrs(summary):
    summary.foo = "bar"
    assert json.load(open("wandb-summary.json")) == {"foo": "bar"}


def test_get_attr(summary):
    assert summary.foo == "init"


def test_update(summary):
    summary.update({"foo": "bar"})
    assert json.load(open("wandb-summary.json")) == {"foo": "bar"}


def test_update_histogram(summary):
    summary.update({"hist": Histogram(np_histogram=([1, 2, 3], [1, 2, 3, 4]))})
    assert json.load(open("wandb-summary.json")) == {
        'foo': 'init',
        "hist": {"_type": "histogram", "values": [1, 2, 3], "bins": [1, 2, 3, 4]}}


def test_set_histogram(summary):
    summary["hist"] = Histogram(np_histogram=([1, 2, 3], [1, 2, 3, 4]))
    assert json.load(open("wandb-summary.json")) == {
        'foo': 'init',
        "hist": {"_type": "histogram", "values": [1, 2, 3], "bins": [1, 2, 3, 4]}}


def test_set_item(summary):
    summary["foo"] = "bar"
    assert json.load(open("wandb-summary.json")) == {"foo": "bar"}


def test_get_item(summary):
    assert summary["foo"] == "init"


def test_delete(summary):
    summary.update({"foo": "bar", "bad": True})
    del summary["bad"]
    assert json.load(open("wandb-summary.json")) == {"foo": "bar"}


def test_big_numpy(summary):
    summary.update({"rad": np.random.rand(1000)})
    assert json.load(open("wandb-summary.json"))["rad"]["max"] > 0
    assert os.path.exists("wandb.h5")


def test_big_nested_numpy(summary):
    summary.update({"rad": {"deep": np.random.rand(1000)}})
    assert json.load(open("wandb-summary.json"))["rad"]["deep"]["max"] > 0
    assert os.path.exists("wandb.h5")


def test_torch_tensor(summary):
    summary.update({"pytorch": torch.rand(1000, 1000)})
    assert os.path.exists("wandb.h5")
    assert json.load(open("wandb-summary.json")
                     )["pytorch"]["_type"] == "torch.Tensor"


def test_tensorflow_tensor(summary):
    with tf.Session().as_default():
        summary.update({"tensorflow": tf.random_normal([1000])})
    assert os.path.exists("wandb.h5")
    assert json.load(open("wandb-summary.json")
                     )["tensorflow"]["_type"] == "tensorflow.Tensor"


def test_pandas(summary):
    summary.update({"pandas": pd.DataFrame(data=np.random.rand(1000))})
    assert os.path.exists("wandb.h5")
    parsed = json.load(open("wandb-summary.json"))
    print(parsed)
    assert parsed["pandas"]["_type"] == "pandas.DataFrame"


def test_read_numpy(summary):
    summary.update({"rad": np.random.rand(1000)})
    s = FileSummary()
    assert len(s["rad"]) == 1000


def test_read_nested_numpy(summary):
    summary.update({"rad": {"deep": np.random.rand(1000)}})
    s = FileSummary()
    assert len(s["rad"]["deep"]) == 1000



import wandb


class Run(object):
    def __init__(self, run=None):
        self.run = run or wandb.run

    def _repr_html_(self):
        url = self.run.get_url()+"?jupyter=true"
        return '''<iframe src="%s" style="border:none;width:100%%;height:420px">
        </iframe>''' % url



import json
import os
import sys
import time
import requests

import wandb
from wandb import util
from wandb.meta import Meta
from wandb.media import Image
from wandb.apis.internal import Api
from gql import gql
import six

SUMMARY_FNAME = 'wandb-summary.json'
DEEP_SUMMARY_FNAME = 'wandb.h5'
H5_TYPES = ("numpy.ndarray", "tensorflow.Tensor",
            "pytorch.Tensor", "pandas.DataFrame")

try:
    import h5py
except ImportError:
    h5py = None


class Summary(object):
    """Used to store summary metrics during and after a run."""

    def __init__(self, **kwargs):
        self._out_dir = kwargs.get("out_dir") or "."
        self._summary = kwargs.get("summary") or {}
        self._h5_path = os.path.join(
            self._out_dir, DEEP_SUMMARY_FNAME)
        # Lazy load the h5 file
        self._h5 = None

    def _write(self, commit=False):
        raise NotImplementedError

    def _transform(self, k, v=None, write=True):
        if not write and isinstance(v, dict):
            if v.get("_type") in H5_TYPES:
                return self.read_h5(k, v)
            else:
                return {key: self._transform(k + "." + key, value, write=False) for (key, value) in v.items()}

        if isinstance(v, wandb.Histogram):
            return wandb.Histogram.transform(v)
        elif isinstance(v, wandb.Graph):
            return wandb.Graph.transform(v)
        else:
            return v

    def __getitem__(self, k):
        return self._transform(k, self._summary[k], write=False)

    def __setitem__(self, k, v):
        self._summary[k] = self._transform(k, v)
        self._write()

    def __setattr__(self, k, v):
        if k.startswith("_"):
            super(Summary, self).__setattr__(k, v)
        else:
            self._summary[k] = self._transform(k, v)
            self._write()

    def __getattr__(self, k):
        if k.startswith("_"):
            return super(Summary, self).__getattr__(k)
        else:
            return self._transform(k, self._summary[k], write=False)

    def __delitem__(self, k):
        val = self._summary[k]
        if isinstance(val, dict) and val.get("_type") in H5_TYPES:
            if not self._h5:
                wandb.termerror("Deleting tensors in summary requires h5py")
            else:
                del self._h5["summary/" + k]
                self._h5.flush()
        del self._summary[k]
        self._write()

    def __repr__(self):
        return str(self._summary)

    def keys(self):
        return self._summary.keys()

    def get(self, k, default=None):
        return self._summary.get(k, default)

    def write_h5(self, key, val):
        # ensure the file is open
        self.open_h5()

        if not self._h5:
            wandb.termerror("Storing tensors in summary requires h5py")
        else:
            try:
                del self._h5["summary/" + key]
            except KeyError:
                pass
            self._h5["summary/" + key] = val
            self._h5.flush()

    def read_h5(self, key, val):
        # ensure the file is open
        self.open_h5()

        if not self._h5:
            wandb.termerror("Reading tensors from summary requires h5py")
        else:
            return self._h5["summary/" + key]

    def open_h5(self):
        if not self._h5 and h5py:
            self._h5 = h5py.File(self._h5_path, 'a', libver='latest')

    def convert_json(self, obj=None, root_path=[]):
        """Convert obj to json, summarizing larger arrays in JSON and storing them in h5."""
        res = {}
        obj = obj or self._summary
        for key, value in six.iteritems(obj):
            path = ".".join(root_path + [key])
            if isinstance(value, dict):
                res[key], converted, transformed = util.json_friendly(
                    self.convert_json(value, root_path + [key]))
            else:
                res[key], converted, transformed = util.json_friendly(value)
                if transformed:
                    if res[key]["_type"] == "pytorch.Tensor":
                        value = value.numpy()
                    elif res[key]["_type"] == "tensorflow.Tensor":
                        value = value.eval()
                    self.write_h5(path, value)
        self._summary = res
        return res

    def update(self, key_vals=None):
        summary = {}
        if key_vals:
            for k, v in six.iteritems(key_vals):
                # TODO: proper image support in summary
                is_image = isinstance(v, Image) or (
                    isinstance(v, list) and isinstance(v[0], Image))
                if is_image or (isinstance(v, dict) and v.get("_type") == "image"):
                    continue
                summary[k] = self._transform(k, v)
        self._summary.update(summary)
        self._write(commit=True)


def download_h5(run, entity=None, project=None, out_dir=None):
    api = Api()
    meta = api.download_url(project or api.settings(
        "project"), DEEP_SUMMARY_FNAME, entity=entity or api.settings("entity"), run=run)
    if meta:
        # TODO: make this non-blocking
        wandb.termlog("Downloading summary data...")
        path, res = api.download_write_file(meta, out_dir=out_dir)
        return path


def upload_h5(file, run, entity=None, project=None):
    api = Api()
    # TODO: unfortunate
    slug = "/".join([project or api.settings("project"), run])
    wandb.termlog("Uploading summary data...")
    api.push(slug, {os.path.basename(file): open(file, 'rb')},
             entity=entity)


class FileSummary(Summary):
    def __init__(self, out_dir="."):
        super(FileSummary, self).__init__(out_dir=out_dir)
        self._fname = os.path.join(out_dir, SUMMARY_FNAME)
        self.load()

    def load(self):
        try:
            self._summary = json.load(open(self._fname))
        except (IOError, ValueError):
            self._summary = {}

    def _write(self, commit=False):
        # TODO: we just ignore commit to ensure backward capability
        with open(self._fname, 'w') as f:
            s = util.json_dumps_safer(self.convert_json(), indent=4)
            f.write(s)
            f.write('\n')
        if self._h5:
            self._h5.close()
            self._h5 = None


class HTTPSummary(Summary):
    def __init__(self, client, run_storage_id, path="//", summary={}, out_dir="."):
        super(HTTPSummary, self).__init__(summary=summary, out_dir=out_dir)
        self._run_storage_id = run_storage_id
        self._path = path
        self._client = client
        self._started = time.time()

    def _write(self, commit=False):
        mutation = gql('''
        mutation UpsertBucket( $id: String, $summaryMetrics: JSONString) {
            upsertBucket(input: { id: $id, summaryMetrics: $summaryMetrics}) {
                bucket { id }
            }
        }
        ''')
        if commit:
            if self._h5:
                self._h5.close()
                self._h5 = None
            res = self._client.execute(mutation, variable_values={
                'id': self._run_storage_id, 'summaryMetrics': util.json_dumps_safer(self.convert_json())})
            assert res['upsertBucket']['bucket']['id']
            entity, project, run = self._path.split("/")
            if os.path.exists(self._h5_path) and os.path.getmtime(self._h5_path) >= self._started:
                upload_h5(self._h5_path, run, entity=entity, project=project)
        else:
            return False



import collections


class Graph(object):
    def __init__(self):
        self.nodes = []

    @classmethod
    def from_keras(cls, model):
        graph = cls()

        # Shamelessly copied from keras/keras/utils/layer_utils.py

        if model.__class__.__name__ == 'Sequential':
            sequential_like = True
        elif not model._is_graph_network:
            # We treat subclassed models as a simple sequence of layers,
            # for logging purposes.
            sequential_like = True
        else:
            sequential_like = True
            nodes_by_depth = model._nodes_by_depth.values()
            nodes = []
            for v in nodes_by_depth:
                if (len(v) > 1) or (len(v) == 1 and len(v[0].inbound_layers) > 1):
                    # if the model has multiple nodes
                    # or if the nodes have multiple inbound_layers
                    # the model is no longer sequential
                    sequential_like = False
                    break
                nodes += v
            if sequential_like:
                # search for shared layers
                for layer in model.layers:
                    flag = False
                    if hasattr(layer, "_inbound_nodes"):
                        for node in layer._inbound_nodes:
                            if node in nodes:
                                if flag:
                                    sequential_like = False
                                    break
                                else:
                                    flag = True
                    if not sequential_like:
                        break

        relevant_nodes = None
        if sequential_like:
            # header names for the different log elements
            to_display = ['Layer (type)', 'Output Shape', 'Param #']
        else:
            relevant_nodes = []
            for v in model._nodes_by_depth.values():
                relevant_nodes += v

        layers = model.layers
        for i in range(len(layers)):
            node = Node.from_keras(layers[i], relevant_nodes)
            graph.nodes.append(node)

        return graph

    @staticmethod
    def transform(graph):
        return {"_type": "graph", "format": "keras", "nodes": [Node.transform(node) for node in graph.nodes]}


class Node(object):
    def __init__(self):
        self.attributes = {}

    @classmethod
    def from_keras(cls, layer, relevant_nodes=None):
        node = cls()

        try:
            output_shape = layer.output_shape
        except AttributeError:
            output_shape = 'multiple'

        node.attributes['name'] = layer.name
        node.attributes['class_name'] = layer.__class__.__name__
        node.attributes['output_shape'] = output_shape
        node.attributes['num_parameters'] = layer.count_params()

        connections = []
        if hasattr(layer, '_inbound_nodes'):
            for in_node in layer._inbound_nodes:
                if relevant_nodes and in_node not in relevant_nodes:
                    # node is not part of the current network
                    continue
                for i in range(len(in_node.inbound_layers)):
                    inbound_layer = in_node.inbound_layers[i].name
                    inbound_node_index = in_node.node_indices[i]
                    inbound_tensor_index = in_node.tensor_indices[i]
                    connections.append(inbound_layer +
                                       '[' + str(inbound_node_index) + '][' +
                                       str(inbound_tensor_index) + ']')
        node.attributes['inbound_nodes'] = connections
        return node

    @staticmethod
    def transform(node):
        return node.attributes


class Histogram(object):
    MAX_LENGTH = 512

    def __init__(self, sequence=None, np_histogram=None, num_bins=64):
        """Accepts a sequence to be converted into a histogram or np_histogram can be set
        to a tuple of (values, bins_edges) as np.histogram returns i.e.

        wandb.log({"histogram": wandb.Histogram(np_histogram=np.histogram(data))})

        The maximum number of bins currently supported is 512
        """
        if np_histogram:
            if len(np_histogram) == 2:
                self.histogram = np_histogram[0]
                self.bins = np_histogram[1]
            else:
                raise ValueError(
                    'Expected np_histogram to be a tuple of (values, bin_edges) or sequence to be specified')
        else:
            try:
                import numpy as np
            except ImportError:
                raise ValueError(
                    "Auto creation of histograms requires numpy")
            self.histogram, self.bins = np.histogram(
                sequence, bins=num_bins)
            self.histogram = self.histogram.tolist()
            self.bins = self.bins.tolist()
        if len(self.histogram) > self.MAX_LENGTH:
            raise ValueError(
                "The maximum length of a histogram is %i" % MAX_LENGTH)
        if len(self.histogram) + 1 != len(self.bins):
            raise ValueError("len(bins) must be len(histogram) + 1")

    @staticmethod
    def transform(histogram):
        return {"_type": "histogram", "values": histogram.histogram, "bins": histogram.bins}



import wandb
import numpy as np
import pytest
import os

data = np.random.randint(255, size=(1000))


def test_raw_data():
    wbhist = wandb.Histogram(data)
    assert len(wbhist.histogram) == 64


def test_np_histogram():
    wbhist = wandb.Histogram(np_histogram=np.histogram(data))
    assert len(wbhist.histogram) == 10


def test_manual_histogram():
    wbhist = wandb.Histogram(np_histogram=([1, 2, 4], [3, 10, 20, 0]))
    assert len(wbhist.histogram) == 3


def test_fucked_up_histogram():
    with pytest.raises(ValueError):
        wbhist = wandb.Histogram(np_histogram=([1, 2, 3], [1]))


def test_transform():
    wbhist = wandb.Histogram(data)
    json = wandb.Histogram.transform(wbhist)
    assert json["_type"] == "histogram"
    assert len(json["values"]) == 64
